!!! hessian.f90 --- 
!! 
!! Description: A collection of routines for calculating the hessian
!!   and gradient of a real function. 
!! 
!!  Some of this code is abstracted from S. Chib / S. Ramamurthy (2010).
!! 
!! Author: Ed Herbst [edward.p.herbst@frb.gov]
!! Last-Updated: 11/20/12
!! 
module hessian 
    use mkl95_precision, only: wp => dp

    use omp_lib
    implicit none 

contains

  subroutine hessp(f, p0, npara, psel, neffpara, f0, grad, hess, info)
    ! Given a real function f of parameter vector p0 (size npara), 
    ! hessp calculates the hessian at p0 for the subset of parameters defined in 
    ! psel (size neffpara.)
    !
    ! The subroutine returns 
    !  - f0: f(p0)
    !  - grad: the neffpara x 1 gradient vector d f(p0)/dp0
    !  - hess: the neffpara x neffpara hessian matrix d2 f(p0)/(dp0dp0')
    !  - info: an integer returning 0 if everything worked.

    real(wp), external :: f 

    integer, intent(in) :: npara, neffpara, psel(neffpara)
    real(wp), intent(in) :: p0(npara)

    integer, intent(out) :: info
    real(wp), intent(out) :: f0, grad(neffpara), hess(neffpara, neffpara)

    real(wp) :: dh(neffpara), p1(npara), grad2(neffpara), grad3(neffpara)
    integer :: i, j

    integer :: nthread

    call gradp(f, p0, npara, psel, neffpara, f0, grad, dh, info)

    if (f0 < -100000000.0_wp) then
       hess = 0.0_wp

       do i = 1, neffpara
          hess(i,i) = -10000.000_wp
       end do
       return
    end if
    !$OMP PARALLEL DO SHARED(p0, grad, hess, psel, neffpara, dh) PRIVATE(i,j,p1)
    do i = 1, neffpara

       do j = i, neffpara
          p1 = p0
          p1(psel(i)) = p1(psel(i)) + dh(i)
          p1(psel(j)) = p1(psel(j)) + dh(j)
          hess(i,j) = ((f(p1) - grad(i)) - grad(j) + f0) / (dh(i)*dh(j))

          hess(j,i) = hess(i, j)
       end do
    end do
    !$OMP END PARALLEL DO

    grad = (grad - f0) / dh
    info = 0
  end subroutine hessp

  subroutine gradp(f, p0, npara, psel, neffpara, f0, grad, dh, info)
    ! grad is NOT the gradient -- its f(p1)
    !--------------------------------------------------------------------------------
    real(wp), external :: f 

    integer, intent(in) :: npara, neffpara, psel(neffpara)
    real(wp), intent(in) :: p0(npara)

    integer, intent(out) :: info
    real(wp), intent(out) :: f0, grad(neffpara), dh(neffpara)


    !---------------------------------------------------
    real(wp), parameter :: ep = 6.0554544523933429e-6

    real(wp) :: p1(npara), xdh

    integer :: i


    f0 = f(p0)

    grad = 0.0_wp
    dh = 0.0_wp

    ! $OMP PARALLEL DO SHARED(p0,grad,psel,dh,neffpara,npara) PRIVATE(i,p1) DEFAULT(shared)
    do i = 1, neffpara

       p1 = p0
       dh(i) = ep*maxval((/abs(p0(psel(i))),0.01_wp/))*(p1(i)+ep)/(abs(p1(i))+ep)

       xdh = p1(i) + dh(i)
       dh(i) = xdh - p1(i)

       p1(psel(i)) = p1(psel(i)) + dh(i)
       grad(i) = f(p1)
       !grad(i) = (f(p1) - f0) / dh(i)

    end do
    ! $OMP END PARALLEL DO
    info = 0

    !write(*,'(a,f)') 'f0 from gradp = ',  f0
  end subroutine gradp


  subroutine cholmod(A, L, indef)
    ! Performs of a modified Cholesky decomposition on A such that L*L'- A is minimized.
    !--------------------------------------------------------------------------------
    real(wp), intent(inout) :: A(:,:)
    real(wp), intent(out) :: L(size(A,1), size(A,1))
    integer, intent(out) :: indef

    real(wp), external :: ddot

    integer :: n, i, j
    real(wp) :: diagA(size(A, 1)), R(size(A,1), size(A,1)), d(size(A,1))
    real(wp) :: gamma, xi, beta, delta, theta, djtemp

    real(wp), allocatable :: ccol(:)

    real(wp) :: REALLY_SMALL


    ! symmetrize
    A = 0.5_wp*(A + transpose(A))
    L = A
    call dpotrf('l',size(A,1), L, size(A,1), indef)

      if (indef == 0) then
         !return
      end if
      L = 0.0_wp
      REALLY_SMALL = epsilon(REALLY_SMALL)

      REALLY_SMALL = 2.2204e-016
      n = size(A, 1)

      R = A
      do i = 1,n
         diagA = A(i,i)
         R(i,i) = 0.0_wp
         L(i,i)=1.0_wp
      end do
      
      gamma = maxval(abs(diagA))
      xi = maxval(abs(R))
      delta = REALLY_SMALL*maxval((/ gamma+xi, 1.0_wp/))
      beta = sqrt(maxval((/gamma, xi/(n*1.0_wp), REALLY_SMALL/)))

      indef = 0

      do j = 1, n
         
         if (j == 1) then
            djtemp = A(j,j)

            if (j < n) then
               allocate(ccol(n-j))
               ccol = A(j+1:n,j)
               theta = maxval(abs(ccol))
               d(j) = maxval((/abs(djtemp), (theta/beta)**2, delta/))
               L(j+1:n, j) = ccol / d(j)
               deallocate(ccol)
            else
               d(j) = maxval((/abs(djtemp), delta/))
            end if
         else

            djtemp = A(j,j) - ddot(j-1, L(j,1:j-1), 1, d(1:j-1) * L(j,1:j-1), 1)

            if (j < n) then
               allocate(ccol(n-j))
               !            print*, 'A = ', A(j+1:n, j)
               ccol = A(j+1:n, j) - matmul(L(j+1:n,1:j-1), d(1:j-1) * L(j,1:j-1))
               !           print*,ccol
               theta = maxval(abs(ccol))
               !            print*,'theta', theta
               d(j) = maxval((/abs(djtemp), (theta/beta)**2, delta/))
               L(j+1:n, j) = ccol / d(j)

               deallocate(ccol)
            else

               d(j) = maxval((/abs(djtemp), delta/))

            end if

         end if

         if (d(j) > djtemp) then

            indef = 1

         end if

      end do

      do j = 1, n

         L(:,j) = L(:,j) * sqrt(d(j))

      end do


    end subroutine cholmod

end module
