!!! mcmc.f90 --- 
!! 
!! Description: A module of MCMC routines.
!! 
!! Author: Ed Herbst [edward.p.herbst@frb.gov]
!! Last-Updated: 09/13/13
!! 
module mcmc 

  use mkl95_precision, only: wp => dp
  use mkl_vsl_type
  use mkl_vsl
  use omp_lib

  use hessian 
  use rand

  implicit none

  integer hits

contains

  subroutine blockmh(para0, M, nsim, f, fh, inbounds, npara, nblocks, pmask, parasim, postsim, acptsim, mumethod, sa_options, & 
       df, sbar, c, bprob, block_burn, max_block_size, blockfile, folder)

    external :: fh
    real(wp), external :: f
    logical, external :: inbounds

    integer, intent(in) :: nsim, npara, df, block_burn, max_block_size
    integer, intent(inout) :: nblocks
    real(wp), intent(in) :: para0(npara), M(npara, npara), sa_options(5), sbar, bprob, c
    real(wp), intent(out) :: parasim(npara, nsim), postsim(nsim) 
    integer, intent(out) :: acptsim(npara, nsim)

    integer :: blocksim(npara,nsim)
    
    character(len=*), intent(in) :: mumethod, folder,blockfile

    character(len=144) :: basedir
    logical :: direxists, good_draw
    integer, intent(in) :: pmask(npara)
    integer :: i, j, b, tot_sim, neffpara
    real(wp) :: f0, f1, p0(npara), p1(npara), alp, q1, q0

    real(wp), allocatable :: muf(:), vf(:,:), R(:,:), rndt(:)
    real(wp) :: s, acptsum(3)

    ! random numbers
    type (VSL_STREAM_STATE) :: stream
    integer errcode
    integer :: brng, method, seed, methodu, time_array(8), mstorage, time_array2(8)
    real(wp) :: u(nsim*npara), eps(nsim*npara), mu(npara), svec(nsim*npara)


    ! blocks
    integer :: bsize, eps_ind, blocksel, ii
    integer, allocatable :: bi(:), break_points(:), ind(:), pmaskinv(:), ind2(:)
    integer :: blocking_type, blockdist(npara,10), ntry

    !timing 
    integer :: start_time, finish_time, cr, cm
    integer, dimension(1) :: st
    real(wp) :: rate

    basedir = trim(folder)
    print*, 'Saving files to:', basedir

    inquire(file=basedir, exist=direxists)
    if (direxists) then
       print *, 'directory already exists'
    else
       call system('mkdir '//basedir)
    endif

    !-------------------------------------------------------------------------
    ! timing and random variables
    !-------------------------------------------------------------------------
    call system_clock(count_rate=cr)
    call system_clock(count_max=cm)
    call system_clock(start_time)
    rate = real(cr)

    call date_and_time(values=time_array)    
    seed = mod(sum(time_array),1000)
    !seed = 425

    brng     = VSL_BRNG_MCG31 !MT19937
    methodu  = VSL_METHOD_DUNIFORM_STD
    method   = VSL_METHOD_DGAUSSIANMV_ICDF
    mstorage = VSL_MATRIX_STORAGE_FULL

    errcode = vslnewstream(stream, brng, seed)
    errcode = vdrnguniform(methodu, stream, nsim*npara, u, 0.0_wp, 1.0_wp)
    errcode = vdrnggaussian(method, stream, nsim*npara, eps, 0.0_wp, 1.0_wp)
    errcode = vdrnguniform(methodu, stream, nsim*npara, svec, 0.0_wp, 1.0_wp)



    !-------------------------------------------------------------------------
    ! blocking
    !-------------------------------------------------------------------------
    neffpara = npara - sum(pmask)

    allocate(pmaskinv(neffpara), ind(neffpara))

    j = 1
    do i = 1, npara
       if (pmask(i) == 0) then
          pmaskinv(j)= i
          j = j + 1;
       end if
    end do

    p0 = para0

    f0 = f(p0)
    print*,'f(p0) = ',f0

    tot_sim = 1
    eps_ind = 0

    if (nblocks == -1) then
       allocate(break_points(neffpara+1))
       blocking_type = 1
    elseif (nblocks == -2) then
       blocking_type = 2

       allocate(ind2(npara),break_points(neffpara+1))
       print*,blockfile 
       open(1, file=blockfile, status='old', action='read')

       ind2 = 0
       break_points = 0
       j = 1
       do i = 1, npara
          if (pmask(i) == 0) then
             read(1,'(i)') ind2(j)
             j = j + 1
          else
             read(1,*)
          end if
       end do
       close(1)

       break_points(1) = 0
       j = 1
       b = 2
       nblocks = 1

       do 
          do i = 1,npara

             if (ind2(i) == nblocks) then
                ind(j) = pmaskinv(i)

                j = j + 1
             end if

          end do

          break_points(b) = j-1

          if (j == neffpara+1) exit 
          nblocks = nblocks + 1
          b = b + 1
       end do

       deallocate(ind2)       


       print*,'nblocks = ', nblocks

    elseif (nblocks == -3) then
       allocate(break_points(npara+1))
       blocking_type = 3
    elseif (nblocks == -4) then
       allocate(break_points(npara+1))
       blocking_type = 4
       print*,blockfile 
       open(1, file=blockfile, status='old', action='read')

       do i = 1,10
          read(1,*) blockdist(:,i)

       end do

       close(1)

    else
       allocate(break_points(nblocks+1))
       blocking_type = 0
    end if
    blocksim = 0 
    acptsim = 0


    call system_clock(st(1))
    call random_seed(PUT=st)
    !---------------------------------------------------------------------------
    ! Begin MCMC sampling
    !---------------------------------------------------------------------------
    do i = 1, nsim

       ind = pmaskinv
       if (blocking_type == 1) then

          call generate_random_sized_blocks(bprob, neffpara, nblocks, ind, break_points)

       elseif (blocking_type == 3) then

          if (i < block_burn) then
             call generate_hessian_blocks(f, p0, npara, neffpara, ind, nblocks, break_points, max_block_size)
          else


             call random_number(s)
             blocksel = floor(s*(block_burn)) + 1
             !             blocksel = 2
             !             print*,'iteration',i,'i`m picking block', blocksel
             !     print*,blocksim(:,blocksel)

             ! call a random integer from 1 to 100, and use those blocks
             break_points = 0
             nblocks = 1
             j = 1
             b = 2
             do 
                do ii = 1,npara

                   if (blocksim(ii, blocksel) == nblocks) then
                      ind(j) = ii!pmaskinv(ii)
                      j = j + 1
                   end if

                end do
                break_points(b) = j - 1

                if (j == neffpara+1) exit
                nblocks = nblocks + 1
                b = b + 1

             end do



          end if

       elseif (blocking_type == 4) then

          call random_number(s)
          blocksel = floor(s*10) + 1


          break_points = 0
          nblocks = 1
          j = 1
          b = 2
          do 
             do ii = 1,npara

                if (blockdist(ii, blocksel) == nblocks) then
                   ind(j) = ii
                   j = j + 1
                end if

             end do
             break_points(b) = j - 1

             if (j == neffpara+1) exit
             nblocks = nblocks + 1
             b = b + 1

          end do

       else

          call generate_random_blocks(bprob, neffpara, nblocks, ind, break_points)

       end if

       do b = 1, nblocks

          
          bsize = break_points(b+1) - break_points(b) 
          allocate(bi(bsize),muf(bsize), vf(bsize, bsize), R(bsize, bsize),rndt(bsize))
          bi = ind(break_points(b)+1:break_points(b+1))
          blocksim(bi,i) = b


          if (mumethod == 'langevin') then
             vf = M(bi,bi)
          end if

          ! Get step size
          s = svec(tot_sim)
          s = sbar*s*c

          call get_proposal_mean_var(f,fh, p0, npara, bsize, bi, muf, vf, R, s, mumethod, sa_options)
          
          p1 = p0

          good_draw = .false.
          ntry = 1
          do 
             if ((good_draw == .true.) .or. ntry > 1) exit
             call vdrngmvt(stream, bsize, rndt, muf, c*R, df)
             p1(bi)= rndt

             !good_draw = inbounds(p1)
             ntry = ntry + 1
          end do

          f1 = f(p1)

          q1 = mvt_pdf(p1(bi), muf, c*R, df)

          if (mumethod .ne. 'sa') then
             call get_proposal_mean_var(f, fh, p1, npara, bsize, bi, muf, vf, R, s, mumethod, sa_options)
          end if

          q0 = mvt_pdf(p0(bi), muf, c*R, df)

          alp = exp(f1 - f0 - q1 + q0)

          if (u(tot_sim) < alp) then
             acptsim(bi, i) = 1
             p0 = p1 
             f0 = f1
          end if

          tot_sim = tot_sim + 1
          deallocate(bi, muf, vf, R,rndt)

          eps_ind = eps_ind + bsize
       end do

       parasim(:,i) = p0

       acptsum(1) = minval((sum(acptsim(pmaskinv,:),2)*1.0_wp)/(1.0_wp*i))
       acptsum(2) = sum(acptsim(pmaskinv,:)*1.0_wp)/(neffpara*i)
       acptsum(3) = maxval((sum(acptsim(pmaskinv,:),2)*1.0_wp)/(1.0_wp*i));

       if (mod(i, 20) == 0) then
          write (*,'(A,I6,A,3F5.2,A,F10.2,A,F4.2)'), "Iteration: ", i, " Acceptance Rate: ", acptsum(:), "   Last Posterior:", f0, " Hits:", (1.0_wp*hits)/(2.0*nblocks*i)
          do b = 1,npara
             print*,b,(sum(acptsim(b,:))*1.0_wp)/(1.0_wp*i)
          end do
       end if

    end do

    call date_and_time(values=time_array2)
    call system_clock(finish_time)
    rate = real(cr)

    open(1,file=trim(basedir)//'/blocksim.txt')
    open(2,file=trim(basedir)//'/acptsim.txt')
    do i = 1,nsim
       write(1,'(100i)') blocksim(:,i)
       write(2,'(100i)') acptsim(:,i)
    end do
    close(1)
    close(2)
    open(1, file=trim(basedir)//'elapsed.txt')
    write(1,'(8i)') time_array
    write(1,'(8i)') time_array2
    close(1)

    deallocate(pmaskinv, ind, break_points)

  end subroutine blockmh

  subroutine get_proposal_mean_var(f, fh, para, npara, neffpara, bpara, muf, vf, R, s,mumethod, sa_options)

    real(wp), external :: f
    external :: fh

    integer, intent(in) :: npara, neffpara, bpara(neffpara)
    real(wp), intent(in) :: para(npara), s, sa_options(5)

    character(len=*), intent(in) :: mumethod
    real(wp), intent(inout) :: muf(neffpara), vf(neffpara, neffpara), R(neffpara, neffpara)

    real(wp) :: grad(neffpara), info_mat(neffpara, neffpara), p1(npara), f0

    integer :: info, ipiv(neffpara),i, j, ntry
    real(wp) :: work(neffpara*neffpara), f1, tr

    real(wp), allocatable :: DG(:,:,:), GDG(:,:,:), GDGG(:,:,:), best_f(:)
    real(wp) :: dh, ep, trj(neffpara), trv(neffpara,1)

    real(wp),external :: ddot
    p1 = para

    ! in this step we compute muf
    select case(mumethod)

    case('newton','newton-sameR')
       call hessp(f, para, npara, bpara, neffpara, f0, grad, info_mat, info)
       info_mat = -info_mat

    case('infomat','infomat-sameR')
       call fh(para, bpara, neffpara, f0, grad, info_mat, info)

    case('riemann')
       ep = 6.0554544523833428e-6
       call fh(para, bpara, neffpara, f0, grad, info_mat, info)
       !      call hessp(f, para, npara, bpara, neffpara, f0, grad, info_mat, info)
       !       info_mat = -info_mat      
       allocate(DG(neffpara, neffpara, neffpara),GDG(neffpara, neffpara, neffpara), GDGG(neffpara, neffpara, neffpara))

       do i = 1, neffpara
          p1 = para
          dh = ep*maxval((/abs(p1(bpara(i))),0.01_wp/))
          p1(bpara(i)) = p1(bpara(i)) + dh

          !          call hessp(f, p1, npara, bpara, neffpara, f0, grad, DG(i,:,:), info)
          call fh(p1, bpara, neffpara, f0, grad, DG(i,:,:), info)

          DG(i,:,:) = (DG(i,:,:) - info_mat)/dh

       end do

       ! vf = inv(hess) -- not necessary for langevin
       call dcopy(neffpara*neffpara, info_mat, 1, vf, 1)
       call dgetrf(neffpara, neffpara, vf, neffpara, ipiv, info)
       call dgetri(neffpara, vf, neffpara, ipiv, work, neffpara*neffpara, info)

       ep = s
       call dcopy(neffpara, para(bpara), 1, muf, 1)
       call dgemv('n', neffpara, neffpara, s**2/2.0_wp, vf, neffpara, grad, 1, 1.0_wp, muf, 1)
       !       print*,para(bpara)
       !       print*,muf
       grad = 0.0_wp
       do i = 1,neffpara
          call dgemm('n','n',neffpara, neffpara, neffpara, 1.0_wp, vf, neffpara, DG(i,:,:),neffpara, 0.0_wp, GDG(i,:,:), neffpara)
          trj(i) = 0.0_wp

          do j = 1,neffpara
             trj(i) = trj(i) + GDG(i,j,j)
          end do

          call dgemm('n','n',neffpara, neffpara, neffpara, 1.0_wp, DG(i,:,:), neffpara, vf, neffpara, 0.0_wp, GDGG(i,:,:),neffpara)

       end do

       do i = 1,neffpara
          do j = 1,neffpara
             muf(i) = muf(i) - ep**2*GDGG(j,i,j)
          end do
       end do
       call dgemv('n',neffpara,neffpara,ep**2/2.0_wp, vf, neffpara, trj, 1, 1.0_wp, muf, 1)
       deallocate(DG,GDG,GDGG)
    case('langevin')
       call gradp(f, para, npara, bpara, neffpara, f0, grad, muf, info)
       grad = (grad - f0)/muf
       call dcopy(neffpara, para(bpara), 1, muf, 1)
       call dgemv('n', neffpara, neffpara, s, vf, neffpara, grad, 1, 1.0_wp, muf, 1)
       call cholmod(vf, R, info)
       return

    case('sa')
       call sa(f, para, npara, bpara, neffpara, f1, p1, sa_options)
       muf = p1(bpara)
       call hessp(f, p1, npara, bpara, neffpara, f0, grad, info_mat, info)
       info_mat = -info_mat


       call dcopy(neffpara*neffpara, info_mat, 1, vf, 1)

       call cholmod(vf, R, info)

       vf = matmul(R,transpose(R))

       call dgetrf(neffpara, neffpara, vf, neffpara, ipiv, info)
       call dgetri(neffpara, vf, neffpara, ipiv, work, neffpara*neffpara, info)

       call dpotrf('l', neffpara, vf, neffpara, info)
       R = vf

       return
    end select

    ! vf = inv(hess) -- not neccesariy for langevin
    !    call dcopy(neffpara, para(bpara), 1, muf, 1)

    !    call dcopy(neffpara*neffpara, info_mat, 1, vf, 1)
    !    call dpotrf('l', neffpara, vf, neffpara, info)
    info = 1
    if (info .ne. 0) then
       call dcopy(neffpara*neffpara, info_mat, 1, vf, 1)
       call dgetrf(neffpara, neffpara, vf, neffpara, ipiv, info)
       call dgetri(neffpara, vf, neffpara, ipiv, work, neffpara*neffpara, info)
       call dgemv('n', neffpara, neffpara, s, vf, neffpara, grad, 1, 0.0_wp, muf, 1)

       ! compute trust ratio
       tr = ddot(neffpara, grad, 1, muf, 1)
       call dgemv('n', neffpara, neffpara, -1.0_wp, info_mat, neffpara, muf, 1, 0.0_wp, trj, 1)

       tr = tr + ddot(neffpara, muf, 1, trj, 1)

       call daxpy(neffpara, 1.0_wp, para(bpara), 1, muf, 1)
    else
       call dpotri('l', neffpara, vf, neffpara, info)
       call dsymv('l', neffpara, s, vf, neffpara, grad, 1, 1.0_wp, muf, 1)
    end if

    !    call fh(p1, bpara, neffpara, f1, grad, info_mat, info)
    if (mumethod=='newton') then
       call hessp(f, p1, npara, bpara, neffpara, f1, grad, info_mat, info)
       info_mat = -info_mat
    elseif (mumethod=='infomat') then
       call fh(p1, bpara, neffpara, f1, grad, info_mat, info)
    elseif (mumethod=='newton-sameR') then
       p1(bpara) = muf
       f1 = f(p1);
!       if (any(bpara==5)) print*,'trust ration = ', (f1 - f0)/tr
    endif


    ! trust region
    if (f1 - f0 < 0.0_wp) then
       muf = para(bpara)

       do i = 1,neffpara
          info_mat(i,i) = 3.0*info_mat(i,i)
       end do
!       if (any(bpara==5)) print*,'bad push'
    else
       hits = hits + 1
    end if

    R = invert_and_cholesky(info_mat, neffpara)


  end subroutine get_proposal_mean_var

  function invert_and_cholesky(matrix, size) result(R)
    ! computes the modifed cholesky
    integer, intent(in) :: size
    real(wp), intent(inout) :: matrix(size,size)

    real(wp) :: R(size, size)

    integer :: info, ipiv(size)
    real(wp) :: work(size*size)

    call cholmod(matrix, R, info)
    call dgemm('n','t',size, size, size, 1.0_wp, R, size, R, size, 0.0_wp, matrix, size)
    call dgetrf(size, size, matrix, size, ipiv, info)
    call dgetri(size, matrix, size, ipiv, work, size*size, info)
    call dpotrf('l', size, matrix, size, info)

    R = matrix

  end function invert_and_cholesky


  subroutine sa(f, para, npara, bpara, neffpara, f1, p1, sa_options)
    ! This code is abstracted from CR 2010
    real(wp), external :: f

    integer, intent(in) :: npara, neffpara, bpara(neffpara)
    real(wp), intent(in) :: sa_options(5), para(npara)

    real(wp), intent(out) :: f1, p1(npara)

    integer :: i, j, ns, nrand, parai, uind, stage_increment, stage_length
    real(wp) :: temp_reduction_factor, temp, sl, p0(npara), arate, cs, scale_factor
    real(wp) :: f0
    integer, allocatable :: acpt(:)

    type (VSL_STREAM_STATE) :: stream
    integer :: method, methodu, status, seed
    real(wp), allocatable :: randu(:), u(:), e(:)

    integer :: time_array(8)

    call date_and_time(values=time_array)
    ns = int(sa_options(1))
    temp = sa_options(2)
    stage_increment = int(sa_options(3))
    temp_reduction_factor = sa_options(4)
    cs = sa_options(5)


    allocate(acpt(ns))
    ! random numbers for select parameter i
    nrand = ns*ns*stage_increment

    allocate(randu(nrand),u(nrand), e(nrand))

    seed=mod(sum(time_array),1000)
    status = vslnewstream(stream, VSL_BRNG_MT19937, seed)
    methodu = VSL_METHOD_DUNIFORM_STD
    method = VSL_METHOD_DGAUSSIANMV_ICDF
    status = vdrnguniform(methodu, stream, nrand, randu, 1.0_wp, neffpara*1.0_wp + 1.0_wp)
    status = vdrnguniform(methodu, stream, nrand, u, 0.0_wp, 1.0_wp)
    status = vdrnggaussian(method, stream, nrand, e, 0.0_wp, 1.0_wp)

    acpt = 0
    stage_length = stage_increment
    uind = 1

    p0 = para
    f0 = f(p0)

    scale_factor = 50

    do i = 1, ns

       do j = 1, stage_length
          p1 = p0

          ! could do validation here
          parai = bpara(floor(randu(uind)))
          
          p1(parai) = p1(parai) + e(uind) / scale_factor;

          f1 = f(p1)

          if (u(uind) < exp((f1 - f0)/temp)) then
             p0 = p1
             f0 = f1

             acpt(i) = acpt(i) + 1

          end if
          uind = uind + 1
       end do

       ! adjust the perturbation scale
       arate = (acpt(i)*1.0_wp) / (stage_length * 1.0_wp)
!       print*,i,arate,scale_factor

       if (arate < 0.0_wp) then
 !         print*,i,arate,uind
       end if
       if (arate  > 0.6_wp) then
          scale_factor = scale_factor / (1.0_wp + cs*(arate - 0.6_wp)/0.4_wp)
       elseif (arate < 0.4_wp) then
          scale_factor = scale_factor * (1.0_wp + cs*(0.4 - arate)/0.4_wp)
       end if

       ! cool the temperature
       temp = temp * temp_reduction_factor
       stage_length = stage_length + stage_increment


    end do

    p1 = p0

    deallocate(randu, u, e, acpt)

  end subroutine sa

  subroutine generate_random_blocks(prob_include, npara, nblocks, indices, break_points)

    integer, intent(in) :: npara, nblocks
    real(wp), intent(in) :: prob_include

    integer, intent(inout) :: indices(npara)
    integer, intent(out) ::  break_points(nblocks+1)

    integer :: ind2(npara)
    integer :: i, gap


    call rperm3(npara, ind2)
    
    indices(ind2) = indices

    gap = int(1.0_wp*npara / nblocks)

    do i = 1, nblocks
       break_points(i) = (i-1)*gap + 1
    end do

    break_points(1) = 0
    break_points(nblocks+1) = npara
    !print*, break_points

  end subroutine generate_random_blocks

  subroutine generate_hessian_blocks(f, para, npara, neffpara, ind, nblocks, break_points, max_block_size)

    real(wp), external :: f

    integer, intent(in) :: npara, neffpara, max_block_size
    integer, intent(inout) :: ind(neffpara)

    integer, intent(out) :: break_points(npara+1), nblocks
    real(wp) :: para(npara)

    integer :: i, j, info, highest_corr, tempi, parai, rem_para, ipiv(neffpara), ii, jj, ii2, jj2, ind2(neffpara)
    real(wp) :: f1, grad(neffpara), hess(neffpara,neffpara), diaghess(neffpara,neffpara), iH(neffpara,neffpara), work(neffpara*neffpara)

    break_points = 0

    call hessp(f, para, npara, ind, neffpara, f1, grad, hess, info)
    hess = -hess

    call dcopy(neffpara*neffpara, hess, 1, iH, 1)
    call dgetrf(neffpara, neffpara, iH, neffpara, ipiv, info)
    call dgetri(neffpara, iH, neffpara, ipiv, work, neffpara*neffpara, info)

!    iH = hess
    diaghess = 0.0_wp
    do i = 1,neffpara
       diaghess(i,i) = 1.0_wp/sqrt(abs(iH(i,i)))
    end do


    ! normalize hessian
    call dgemm('n','n',neffpara, neffpara, neffpara, 1.0_wp, diaghess, neffpara, iH, neffpara, 0.0_wp, hess, neffpara)
    call dgemm('n','n',neffpara, neffpara, neffpara, 1.0_wp, hess, neffpara, diaghess, neffpara, 0.0_wp, iH, neffpara)

    iH = abs(iH)
    do j = 1,neffpara
       iH(j,j) = -10000.0_wp
    end do

    rem_para = neffpara

    j = 1

    nblocks = 0
    break_points = 0

    ind2 = ind
    do while (rem_para > 0) 
       
       jj = maxloc(maxval(iH(:,:),dim=1),dim=1)
       ii = maxloc(iH(:,jj),dim=1)

!       if (any(ind(1:j-1)==ii) .or. any(ind(1:j-1)==jj)) then
!          print*,ii,jj,iH(ii,jj)
 !         stop
!       end if
       ind(j) = ind2(ii)
       ind(j+1) = ind2(jj)

       j = j + 2

       iH(ii,jj) = -1000000.0_wp
       iH(jj,ii) = -1000000.0_wp
       rem_para = rem_para - 2

       if ((max_block_size > 2) .and. (rem_para > 0)) then
          !jj2 = maxloc(iH(:,jj),dim=1)
          ii2 = maxloc(iH(:,ii),dim=1)

          ind(j) = ind2(ii2)
          
          j = j + 1
          rem_para = rem_para - 1
          
          iH(ii2,:) = -1000000.0_wp
          iH(:,ii2) = -1000000.0_wp
       end if

       iH(ii, :) = -1000000.0_wp
       iH(:, ii) = -1000000.0_wp

       iH(jj, :) = -1000000.0_wp
       iH(:, jj) = -1000000.0_wp
       
       if ((ii == 41) .or. (jj == 41)) then
!          print*, ii,jj,ii2
       end if

       nblocks = nblocks + 1
       break_points(nblocks+1) = j - 1;

       if ((rem_para < max_block_size) .and. (rem_para > 0)) then

          do ii = j,neffpara
             jj = maxloc(maxval(iH(:,:),dim=1),dim=1)

             ind(ii) = ind2(jj)
             iH(:,jj) = -1000000.0_wp
             iH(jj,:) = -1000000.0_wp
             
          end do

          rem_para = 0
          nblocks = nblocks + 1
       end if
    end do

    break_points(nblocks+1) = neffpara

!!$    do while (rem_para > 0) 
!!$
!!$       parai = ind(j)
!!$
!!$       highest_corr = maxloc(iH(ind(j+1:neffpara),parai),1)
!!$
!!$       tempi = ind(j+1)
!!$
 !!$       ind(j+1) = ind(j+highest_corr)
!!$       ind(j+highest_corr) = tempi
!!$
!!$       j = j + 2
!!$       if (j >= neffpara-1) then
!!$          rem_para = 0
!!$       end if
!!$
!!$       nblocks = nblocks + 1
!!$       break_points(nblocks) = j - 1
!!$    end do

!    break_points(nblocks+1) = neffpara

  end subroutine generate_hessian_blocks

  subroutine generate_random_sized_blocks(prob_include, npara, nblocks, indices, break_points)

    integer, intent(in) :: npara
    integer, intent(inout) :: nblocks
    real(wp), intent(in) :: prob_include

    integer, intent(inout) :: indices(npara)
    integer, intent(out) ::  break_points(npara+1)

    real(wp) :: u
    integer :: i, j, ind2(npara)
    call rperm3(npara, ind2)


    indices(ind2) = indices

    break_points = 0
    break_points(1) = 1
    j = 2
    do i = 2, npara

       call random_number(u)

       if (u < (1.0_wp - prob_include)) then
          break_points(j) = i

          j = j + 1
       end if
    end do

    break_points(j) = npara

    if (break_points(j) == break_points(j-1)) then
       nblocks = j-2
    else
       nblocks = j-1
    end if


  end subroutine generate_random_sized_blocks

  subroutine rperm3(N, p)

    integer, intent(in) :: N
    integer, dimension(N), intent(out) :: p

    integer :: j, k
    real(wp) :: u

    p = 0

    do j=1,N

       call random_number(u)
       k = floor(j*u) + 1

       p(j) = p(k)
       p(k) = j

    end do

  end subroutine rperm3

  subroutine rwmh(para0, proposal_var, nsim, f, npara, pmask, parasim, postsim, acptsim, nblocks, blockfile, folder)
    external f
    !f2py call f()
    real(wp) :: f
    integer, intent(in) :: nsim, npara, pmask(npara)
    real(wp), intent(in) :: para0(npara), proposal_var(npara, npara)


    real(wp), intent(out) :: parasim(npara, nsim), postsim(nsim) 
    integer, intent(out) :: acptsim(nsim)

    integer, intent(inout) :: nblocks
    ! mask
    integer :: neffpara
    integer, allocatable :: pmaskinv(:)

    ! saving files
    character(len=144), intent(in) :: folder, blockfile
    character(len=144) :: basedir, parafile, postfile, acptfile
    logical :: direxists
    integer :: i, j
    real(wp) :: f0, f1, p0(npara), p1(npara), alp

    ! random numbers
    type (VSL_STREAM_STATE) :: stream
    integer :: errcode, info
    integer :: brng, method, seed, methodu, time_array(8), mstorage, time_array2(8)
    real(wp) :: u(nsim*npara), s
    real(wp), allocatable :: eps(:), mu(:), masked_var(:,:), eye(:,:), muf(:)

    integer :: bsize, epsind, b, uind
    integer, allocatable :: bi(:), break_points(:), ind(:), ind2(:)
    integer :: blocking_type, blockdist(npara,10), ii, blocksel
    real(wp), allocatable :: R(:,:)

    !timing 
    integer :: start_time, finish_time, cr, cm
    real(wp) :: rate

    ! files stuff
    !---------------------------------------------------------------------
    !print*, 'Saving files to:', basedir

!!$    inquire(file=basedir, exist=direxists)
!!$    if (direxists) then
!!$       print *, 'directory already exists'
!!$    else
!!$       call system('mkdir '//basedir)
!!$    endif
    !---------------------------------------------------------------------

    basedir = folder

    ! start time
    call system_clock(count_rate=cr)
    call system_clock(count_max=cm)
    call system_clock(start_time)
    rate = real(cr)
    ! mask stuff
    !---------------------------------------------------------------------
    neffpara = npara - sum(pmask)
    allocate(pmaskinv(neffpara), eps(nsim*npara), mu(neffpara), masked_var(neffpara,neffpara), eye(neffpara, neffpara),ind(neffpara))
    pmaskinv = 0
    eye = 0.0_wp
    j = 1
    do i =  1, npara
       if (pmask(i) == 0) then
          pmaskinv(j) = i
          eye(j,j) = 1.0_wp
          j = j+1
       endif
    end do
    !---------------------------------------------------------------------

    !random number garbage
    !---------------------------------------------------------------------
    mu = 0.0_wp
    call date_and_time(values=time_array)

    brng=VSL_BRNG_MCG31 !MT19937
    seed=mod(sum(time_array),1000)
    print*,'RWMH seed: ', seed
    methodu = VSL_METHOD_DUNIFORM_STD
    method = VSL_METHOD_DGAUSSIANMV_ICDF
    mstorage = VSL_MATRIX_STORAGE_FULL

    errcode = vslnewstream(stream, brng, seed)
    errcode = vdrnguniform(methodu, stream, nsim*npara, u, 0.0_wp, 1.0_wp)

    ! we are now accepting the proposal variance, we need to do a cholesky decomposition
    masked_var = proposal_var(pmaskinv, pmaskinv)    
    call dpotrf('l',neffpara, masked_var, neffpara, info)
    print*,'variance info', info
    !    errcode = vdrnggaussianmv(method, stream, nsim, eps, npara, mstorage, mu, eye)
    errcode = vdrnggaussian(method, stream, nsim*npara, eps, 0.0_wp, 1.0_wp)

    if (nblocks == -1) then
       allocate(break_points(neffpara+1))
       blocking_type = 1
    elseif (nblocks == -2) then
       blocking_type = 2
       allocate(ind2(npara),break_points(neffpara+1))
       print*,blockfile 
       open(1, file=blockfile, status='old', action='read')

       ind2 = 0
       break_points = 0
       j = 1
       do i = 1, npara
          if (pmask(i) == 0) then
             read(1,'(i)') ind2(j)
             j = j + 1
          else
             read(1,*)
          end if
       end do
       close(1)

       break_points(1) = 0
       j = 1
       b = 2
       nblocks = 1

       do 
          do i = 1,npara

             if (ind2(i) == nblocks) then
                ind(j) = pmaskinv(i)

                j = j + 1
             end if

          end do

          break_points(b) = j-1

          if (j == neffpara+1) exit 
          nblocks = nblocks + 1
          b = b + 1
       end do

       deallocate(ind2)       


       print*,'nblocks = ', nblocks

    elseif (nblocks == -4) then
       allocate(break_points(npara+1))
       blocking_type = 4
       print*,blockfile 
       open(1, file=blockfile, status='old', action='read')

       do i = 1,10
          read(1,*) blockdist(:,i)

       end do

       close(1)

    else
       allocate(break_points(nblocks+1))
       blocking_type = 0
    end if


    !---------------------------------------------------------------------
    acptsim = 0

    p0 = para0
    f0 = f(p0)
    uind = 1

    epsind = 0;
    do i = 1, nsim


       if (blocking_type == 1) then
          ind = pmaskinv
          call generate_random_sized_blocks(0.50_wp, neffpara, nblocks, ind, break_points)
       elseif (blocking_type == 0) then
          ind = pmaskinv
          call generate_random_blocks(0.50_wp, neffpara, nblocks, ind, break_points)
       elseif (blocking_type == 4) then
                    call random_number(s)
          blocksel = floor(s*(10-1)) + 1

          ! call a random integer from 1 to 100, and use those blocks
          break_points = 0
          nblocks = 1
          j = 1
          b = 2
          do 
             do ii = 1,npara

                if (blockdist(ii, blocksel) == nblocks) then
                   ind(j) = ii!pmaskinv(ii)
                   j = j + 1
                end if

             end do
             break_points(b) = j - 1

             if (j == neffpara+1) exit
             nblocks = nblocks + 1
             b = b + 1

          end do

       end if

       do b = 1, nblocks

          bsize = break_points(b+1) - break_points(b) 
          allocate(bi(bsize), R(bsize, bsize),muf(bsize))
          bi = ind(break_points(b)+1:break_points(b+1))

          R = proposal_var(bi,bi)
          call dpotrf('l', bsize, R, bsize, info)

          if (info .ne. 0) then
             print*,'Cholesky failed.', info
          end if

          p1 = p0
          muf = p1(bi)
!          p1(bi) = p0(bi) + matmul(R,eps(epsind+1:epsind+bsize))

          call dsymv('l', bsize, 1.0_wp, R, bsize, eps(epsind+1:epsind+bsize), 1, 1.0_wp, muf, 1)

          p1(bi)=muf

          f1 = f(p1)

          alp = exp(f1 - f0)

          if (u(uind) < alp) then
             acptsim(i) = 1
             p0 = p1 
             f0 = f1
          end if

!          parasim(bi, i) = p0

          deallocate(bi, R, muf)

          epsind = epsind + bsize
          uind = uind + 1

       end do
       parasim(:,i) = p0
       postsim(i) = f0

       if (mod(i, 1000) == 0) then
          write (*,'(A,I6,A,F4.2,A,F10.2)'), "Iteration: ", i, " Acceptance Rate: ", sum(acptsim*1.0_wp)/i, "   Last Posterior:", f0
       end if

    end do

    parafile = trim(basedir)//'/'//'parasim.txt'
    postfile = trim(basedir)//'/'//'postsim.txt'
    acptfile = trim(basedir)//'/'//'acptsim.txt'

    open(1, file=parafile, action='write')
    open(2, file=postfile, action='write')
    open(3, file=acptfile, action='write')

    do i = 1, nsim
       write(1, '(100f)') parasim(:,i)
       write(2, '(1f)') postsim(i)
       write(3, '(1i)') acptsim(i)
    end do

    close(1)
    close(2)
    close(3)
    
    print *,'Wrote files {parasim,postsim,acptsim}.txt'


    call system_clock(finish_time)
    rate = real(cr)
    call date_and_time(values=time_array2)
    open(1, file=trim(basedir)//'/elapsed.txt');
!    write(1,*) (finish_time - start_time)/rate
    write(1,'(8i)') time_array
    write(1,'(8i)') time_array2
    close(1)

    deallocate(pmaskinv, eps, mu, masked_var, eye, break_points, ind)
  end subroutine rwmh

!!$  subroutine smc(f, g, phi, nphi, npart, npara, nintmh, parasim, wtsim, pmask, folder, kapp, nblocks, h)
!!$
!!$    real(wp), external :: f, g
!!$    real(wp), external, optional :: h
!!$
!!$    integer, intent(in) :: nphi, npart, npara, nintmh, pmask(npara), nblocks
!!$
!!$    real(wp), intent(inout) :: parasim(npara, npart), wtsim(npart), phi(nphi)
!!$    character(*), intent(in) :: folder
!!$    real(wp), intent(in) :: kapp
!!$
!!$    integer :: neffpara
!!$    integer, allocatable :: pmaskinv(:)
!!$    real(wp), allocatable :: eye(:,:)
!!$    real(wp) :: lik(npart), ESS(nphi), var(npara, npara), R(npara, npara), scale
!!$    real(wp) :: mu(npara), prio(npart), incwt(npart), wtsq(npart), zi(nphi)
!!$
!!$    real(wp) :: p0(npara), p1(npara), lik0, lik1, pr0, pr1, alp, arate(nphi,npara)
!!$    real(wp) :: parasimcopy(npara, npart), m(npart), cs, cweights(npart), uu(nphi)
!!$
!!$    real(wp) :: q0, q1, h0, h1
!!$    integer :: i, j, k, paraind(npart)
!!$    integer :: modi, uind
!!$    integer :: info
!!$    integer :: acptsim(npart, npara)
!!$
!!$    character(len=144) :: datafile, chari, charnpara, phifile, wghtfile, zifile, llikfile, postfile
!!$    logical :: direxists
!!$
!!$    ! random numbers
!!$    type (VSL_STREAM_STATE) :: stream
!!$    integer errcode
!!$    integer :: brng, method, seed, methodu, time_array(8), mstorage
!!$    real(wp) :: u(npart*nintmh*nphi*nblocks)
!!$    real(wp), allocatable :: eps(:,:)
!!$
!!$    real(wp) :: mixr
!!$
!!$
!!$    ! block stuff
!!$    integer :: bj, bsize
!!$    integer, allocatable :: bi(:), break_points(:), ind(:)
!!$    real(wp), allocatable :: bvar(:,:)
!!$    ! file i/o
!!$    !---------------------------------------------------------------------
!!$    print*, 'Saving files to:', folder
!!$
!!$    inquire(file=folder, exist=direxists)
!!$    if (direxists) then
!!$       print *, 'directory already exists'
!!$    else
!!$       call system('mkdir '//folder)
!!$    endif
!!$    phifile = trim(folder)//'/phi.txt'
!!$
!!$    open(1, file=phifile, action='write')
!!$    do i = 1, nphi
!!$       write(1, '(1f)') phi(i)
!!$    end do
!!$    close(1)
!!$    !---------------------------------------------------------------------
!!$    neffpara = npara - sum(pmask)
!!$    allocate(pmaskinv(neffpara), ind(neffpara), eye(neffpara, neffpara), eps(nintmh*npart, npara))
!!$    !pmaskinv = 1 - pmask
!!$    eye = 0.0_wp
!!$    j = 1
!!$    do i = 1, npara
!!$       if (pmask(i) == 0) then
!!$          pmaskinv(j) = i
!!$          eye(j,j) = 1.0_wp
!!$          j = j + 1
!!$       end if
!!$    end do
!!$    !print*, '# eff para:', pmaskinv
!!$
!!$    mu = 0.0_wp
!!$    call date_and_time(values=time_array)
!!$
!!$    brng=VSL_BRNG_MT19937
!!$    seed=mod(sum(time_array),1000)
!!$    print*,seed, kapp
!!$    !seed = 425
!!$    methodu=VSL_METHOD_DUNIFORM_STD
!!$    method=VSL_METHOD_DGAUSSIANMV_ICDF
!!$    mstorage=VSL_MATRIX_STORAGE_FULL
!!$
!!$    errcode = vslnewstream(stream, brng, seed)
!!$    errcode = vdrnguniform(methodu, stream, nphi, uu, 0.0_wp, 1.0_wp)
!!$    errcode = vdrnguniform(methodu, stream, npart*nphi*nintmh*nblocks, u, 0.0_wp, 1.0_wp)
!!$
!!$    scale = 0.2_wp
!!$    arate(1,:) = 1.0_wp
!!$    modi = 100
!!$
!!$    allocate(break_points(nblocks+1))
!!$    write(*,'(a$)') 'Evaluating initial likelihoods. '
!!$    do i = 1, npart
!!$       lik(i) = f(parasim(:,i))
!!$       prio(i) = g(parasim(:,i))
!!$       if (mod(i, modi) == 0) then
!!$          write(*,'(a$)') '.'
!!$       endif
!!$
!!$    end do
!!$    write(*,'(a,f,a)') ' finished. (mean unscaled lik', sum(lik)/(1.0_wp*npart), ')'
!!$
!!$    wtsim =  exp(prio - wtsim)
!!$
!!$    wtsim = wtsim / sum(wtsim)
!!$    ESS(1) = 1.0_wp / sum(wtsim**2)
!!$
!!$    ! this should be changed for non exact simulations at i=1
!!$    zi(1) = 1.0_wp 
!!$
!!$    print*,'ESS(1) = ', ESS(1)
!!$
!!$    write(charnpara,'(i2)') npara
!!$    write(chari,'(i3.3)') i
!!$
!!$    datafile = trim(folder)//'/'//trim(chari)//'parasim.txt'
!!$    wghtfile = trim(folder)//'/'//trim(chari)//'wtsim.txt'
!!$
!!$    open(1, file=datafile, action='write')
!!$    open(2, file=wghtfile, action='write')
!!$
!!$    do j = 1, npart
!!$
!!$       write(1,'(41f)') parasim(:,j)
!!$       write(2,'(1f)') wtsim(j)
!!$
!!$    end do
!!$
!!$    close(1)
!!$    close(2)
!!$
!!$    write(*,*) 'Wrote files '//trim(chari)//'{parasim,wtsim,ess}.txt'
!!$    print*,''
!!$
!!$    zifile = trim(folder)//'/zratios.txt'
!!$    open(3, file=zifile, action='write')
!!$    write(3, '(f)') zi(1)
!!$
!!$    uind = 1
!!$    do i = 2, nphi
!!$       call vdexp(npart, (phi(i) - phi(i-1))*lik, incwt)
!!$       !call vsmul(npart, wtsim, incwt, wtsim)
!!$
!!$       wtsim = wtsim * incwt
!!$       zi(i) = sum(wtsim)
!!$       write(3, '(f)') zi(i)
!!$
!!$       wtsim = wtsim / zi(i)
!!$       call vdsqr(npart, wtsim, wtsq)
!!$
!!$       ESS(i) = 1.0_wp/sum(wtsq)
!!$       write(*,'(a,f4.2,a,f$)') 'phi[', phi(i), '] -- [ESS = ', ESS(i)
!!$
!!$       if (ESS(i) < npart/2.0) then 
!!$          write (*, '(a$)') ' resample]'
!!$          call systematic_resampling(npart, wtsim, uu(i), paraind)
!!$
!!$          parasim = parasim(:, paraind)
!!$          lik = lik(paraind)
!!$          prio = prio(paraind)
!!$
!!$          wtsim = 1.0_wp / npart
!!$
!!$       else
!!$
!!$          write (*, '(a$)') '         ]'
!!$
!!$       end if
!!$
!!$       call compute_mean_and_variance(mu, var, parasim, wtsim, npara, npart)
!!$
!!$       mu = 0.0_wp
!!$
!!$       errcode = vdrnggaussianmv(method, stream, npart*nintmh, eps, npara, & 
!!$            mstorage, mu, scale*eye)
!!$       
!!$       acptsim = 0
!!$
!!$       ind = pmaskinv
!!$       call generate_random_blocks(0.0_wp, neffpara, nblocks, ind, break_points)
!!$
!!$       !$OMP PARALLEL DO SHARED(npart,ind, u,eps,acptsim, parasim, lik, prio, nblocks, kapp, nintmh,i,scale,var) PRIVATE(j, bvar, bj, bsize, bi, k, p1, p0, lik0, pr0, mixr, q0, q1, h0, h1, pr1, lik1,alp)
!!$       do j = 1, npart
!!$          p0 = parasim(:,j)
!!$          lik0 = lik(j)
!!$          pr0 = prio(j)
!!$
!!$          do bj = 1, nblocks
!!$             bsize = break_points(bj+1) - break_points(bj) 
!!$
!!$             allocate(bi(bsize), bvar(bsize, bsize))
!!$
!!$             bi = ind(break_points(bj)+1:break_points(bj+1))
!!$
!!$             bvar = var(bi, bi)
!!$             call dpotrf('l', bsize, bvar, bsize, info)
!!$
!!$             if (info .ne. 0) then
!!$                print*,'Cholesky failed.', info
!!$                var = 0.0_wp
!!$                do k = 1, npara
!!$                   var(k,k) = 0.01_wp
!!$                end do
!!$
!!$             end if
!!$
!!$             !print*, 'now on block', bj,'ind', bi
!!$             do k = 1, nintmh
!!$                p1 = p0
!!$
!!$                ! temporary test 
!!$                call random_number(mixr)
!!$
!!$                q0 = 0.0_wp
!!$                q1 = 0.0_wp
!!$
!!$                h0 = 0.0_wp
!!$                h1 = 0.0_wp
!!$
!!$                if (mixr < kapp) then
!!$                   p1(bi) = p1(bi) + matmul(bvar, eps((j-1)*nintmh + k,bi))
!!$                elseif (mixr < (kapp + (1.0_wp - kapp)/2.0_wp)) then
!!$                   p1(bi) = p1(bi) + 3.0_wp*matmul(bvar, eps((j-1)*nintmh + k,bi))
!!$                else
!!$                   p1(bi) = mu(bi) + matmul(bvar, eps((j-1)*nintmh + k,bi))
!!$                   q0 = mvnormal_pdf(p0(bi), mu(bi), scale*bvar) 
!!$                   q1 = mvnormal_pdf(p1(bi), mu(bi), scale*bvar) 
!!$                endif
!!$
!!$                pr1 = g(p1)
!!$                lik1 = f(p1)
!!$
!!$                if (present(h)) then
!!$                   h1 = h(p1)
!!$                   h0 = h(p0)
!!$                end if
!!$
!!$                alp = exp(phi(i)*(lik1 - lik0) + pr1 - pr0 + q0 - q1 + h1 - h0)
!!$
!!$                if (u((i-2)*nintmh*npart*nblocks + (j-1)*nintmh*nblocks + (bj-1)*nintmh + k) < alp) then
!!$
!!$                   p0 = p1
!!$                   lik0 = lik1
!!$                   pr0 = pr1
!!$
!!$                   acptsim(j,bi) = acptsim(j,bi) + 1
!!$
!!$                end if
!!$
!!$             end do ! k = 1, nintmh
!!$
!!$             parasim(:,j) = p0
!!$             lik(j)  = lik0
!!$             prio(j) = pr0
!!$
!!$             deallocate(bi, bvar)
!!$          end do ! blocks
!!$       end do
!!$       !$OMP END PARALLEL DO
!!$
!!$       arate(i,:) = sum(acptsim, dim=1)/(1.0_wp*npart*nintmh)!
!!$       print*,''
!!$       write(*,'(a,f5.2,f5.2,f5.2)') ' acceptance rate [min,mean,max]: ', & 
!!$            minval(arate(i,pmaskinv)), sum(arate(i,pmaskinv))/neffpara, maxval(arate(i,pmaskinv))
!!$       write(*,'(a,f,f)'), '(unweighted) average likelihood + posterior: ', &
!!$            sum(lik)/(npart*1.0_wp), sum(lik+prio)/(npart*1.0_wp)
!!$       print*,sum(lik)/(npart*1.0_wp)
!!$       print*, 'mdd estimate', sum(log(zi(1:i)))
!!$       !------------------------------------------------------------
!!$       write(charnpara,'(i3)') npara
!!$       write(chari,'(i3.3)') i
!!$
!!$       datafile = trim(folder)//'/'//trim(chari)//'parasim.txt'
!!$       wghtfile = trim(folder)//'/'//trim(chari)//'wtsim.txt'
!!$       llikfile = trim(folder)//'/'//trim(chari)//'liksim.txt'
!!$       postfile = trim(folder)//'/'//trim(chari)//'postsim.txt'
!!$       open(1, file=datafile, action='write')
!!$       open(2, file=wghtfile, action='write')
!!$       open(4, file=llikfile, action='write')
!!$       open(5, file=postfile, action='write')
!!$       do j = 1, npart
!!$
!!$          write(1,'(100f)') parasim(:,j)
!!$          write(2,'(1f)') wtsim(j)
!!$          write(4,'(1f)') lik(j)
!!$          write(5,'(1f)') lik(j)+prio(j)
!!$
!!$       end do
!!$
!!$       close(1)
!!$       close(2)
!!$       close(4)
!!$       close(5)
!!$       !----------------------------------------------------------
!!$       if (minval(arate(i,pmaskinv)) < 0.25_wp) then
!!$          scale = maxval((/ scale - 0.01_wp, 0.01_wp /))
!!$          print*,'adjusting R down: ', scale
!!$       else
!!$          scale = minval((/ scale + 0.01_wp, 1.00_wp/))
!!$          print*,'adjusting R up: ', scale
!!$       end if
!!$
!!$       write(*,*) 'Wrote files '//trim(chari)//'{parasim,wtsim,liksim}.txt'
!!$       print*,''
!!$    end do
!!$    close(3)
!!$    deallocate(pmaskinv, break_points, ind, eye, eps)
!!$  end subroutine smc
!!$
!!$
!!$  subroutine gewekes_smc(f, g, T, nt, npart, npara, nintmh, parasim, wtsim, pmask, folder, nblocks)
!!$
!!$    real(wp), external :: f, g
!!$    integer, intent(in) :: nt, npart, npara, nintmh, pmask(npara), nblocks
!!$
!!$    real(wp), intent(inout) :: parasim(npara, npart), wtsim(npart)
!!$    integer :: T(nt)
!!$
!!$    character(*), intent(in) :: folder
!!$
!!$    integer :: neffpara
!!$    integer, allocatable :: pmaskinv(:)
!!$
!!$    real(wp) :: likp(npart), lik(npart), ESS(nT), var(npara, npara), R(npara, npara), scale
!!$    real(wp) :: mu(npara), prio(npart), incwt(npart), wtsq(npart)
!!$    real(wp) :: p0(npara), p1(npara), lik0, lik1, pr0, pr1, alp, arate(nT,npara)
!!$    real(wp) :: parasimcopy(npara, npart), m(npart), cs, cweights(npart), uu(nT)
!!$    real(wp) :: even_wt(npart)
!!$
!!$    integer :: i, j, k, paraind(npart), ii
!!$    integer :: modi
!!$    integer :: info
!!$    integer :: acptsim(npart,npara)
!!$
!!$    character(len=144) :: datafile, basedir, chari, charnpara, phifile, wghtfile
!!$    logical :: direxists
!!$    ! random numbers
!!$    type (VSL_STREAM_STATE) :: stream
!!$    integer errcode
!!$    integer :: brng, method, seed, methodu, time_array(8), mstorage
!!$    real(wp) :: u(npart*nT*nintmh*nblocks), eps(npart*nintmh, npara), eye(npara, npara)
!!$
!!$    ! block stuff
!!$    integer :: bj, bsize
!!$    integer, allocatable :: bi(:), break_points(:), ind(:)
!!$    real(wp), allocatable :: bvar(:,:)
!!$
!!$    ! file i/o
!!$    !-------------------------------------------------------------
!!$    print*, 'Saving files to:', folder
!!$
!!$    inquire(file=folder, exist=direxists)
!!$    if (direxists) then
!!$       print *, 'directory already exists'
!!$    else
!!$       call system('mkdir -p '//folder)
!!$    endif
!!$    basedir = folder
!!$    phifile = trim(folder)//'/T.txt'
!!$
!!$    open(1, file=phifile, action='write')
!!$    do i = 1, nT
!!$       write(1, '(1i)') T(i)
!!$    end do
!!$    close(1)
!!$    !--------------------------------------------------------------
!!$    even_wt = 1.0_wp / npart
!!$    neffpara = npara - sum(pmask)
!!$    allocate(pmaskinv(neffpara), ind(neffpara), break_points(nblocks+1))
!!$    j = 1
!!$    do i = 1, npara
!!$       eye(i,i)= 1.0_wp
!!$       if (pmask(i) == 0) then
!!$          pmaskinv(j) = i
!!$          j = j + 1
!!$       end if
!!$    end do
!!$
!!$    mu = 0.0_wp
!!$    call date_and_time(values=time_array)
!!$
!!$    brng=VSL_BRNG_MCG31 !MT19937
!!$    seed=mod(sum(time_array),1000)
!!$    print*,seed
!!$    !seed = 425
!!$    methodu=VSL_METHOD_DUNIFORM_STD
!!$    method=VSL_METHOD_DGAUSSIANMV_ICDF
!!$    mstorage=VSL_MATRIX_STORAGE_FULL
!!$
!!$    errcode = vslnewstream(stream, brng, seed)
!!$    errcode = vdrnguniform(methodu, stream, nT, uu, 0.0_wp, 1.0_wp)
!!$    errcode = vdrnguniform(methodu, stream, npart*nT*nintmh*nblocks, u, 0.0_wp, 1.0_wp)
!!$
!!$    scale = 0.2_wp
!!$    arate(1,:) = 1.0_wp
!!$    modi = 100
!!$    write(*,'(a$)') 'Evaluating initial likelihoods. '
!!$    do i = 1, npart
!!$       lik(i) = f(parasim(:,i), T(1))
!!$       likp(i) = f(parasim(:,i), T(2))
!!$       prio(i) = g(parasim(:,i))
!!$
!!$       if (mod(i, modi) == 0) then
!!$          write(*,'(a$)') '.'
!!$       endif
!!$
!!$    end do
!!$    write(*,'(a,f,a)') ' finished. (mean scaled lik', sum(lik)/(1.0_wp*npart), ')'
!!$
!!$    wtsim = exp(prio - wtsim)
!!$    ESS(1) = 1.0_wp / sum(wtsim**2)
!!$    print*,ESS(1)
!!$    do i = 2, nT
!!$
!!$       call vdexp(npart, likp-lik, incwt)
!!$       !call vsmul(npart, wtsim, incwt, wtsim)
!!$
!!$       !print*,'previous ESS', 1.0_wp/sum(wtsim**2)
!!$       wtsim = wtsim * incwt
!!$       wtsim = wtsim / sum(wtsim)
!!$       call vdsqr(npart, wtsim, wtsq)
!!$
!!$       ESS(i) = 1.0_wp/sum(wtsim**2)
!!$       write(*,'(a,i,a,f$)') 'N[', T(i), '] -- [ESS = ', ESS(i)
!!$
!!$       if (ESS(i) < npart/2.0_wp) then 
!!$          write (*, '(a$)') ' resample]'
!!$
!!$          call systematic_resampling(npart, wtsim, uu(i), paraind)
!!$
!!$          wtsim = 1.0_wp / npart
!!$
!!$          parasim = parasim(:, paraind)
!!$
!!$          ! do j = 1, npart
!!$          !   wtsim(j) = exp(f(parasim(:,j), T(i)))
!!$          ! end do
!!$          ! wtsim = wtsim / sum(wtsim)
!!$          ! print*,'Actual ESS = ' , 1.0_wp/sum(wtsim**2) 
!!$       else
!!$          write (*, '(a$)') '         ]'
!!$       end if
!!$
!!$       mu = 0.0_wp
!!$
!!$       errcode = vdrnggaussianmv(method, stream, npart*nintmh, eps, npara, & 
!!$            mstorage, mu, scale*eye)
!!$
!!$       call compute_mean_and_variance(mu, var, parasim, wtsim, npara, npart)
!!$
!!$       ind = pmaskinv
!!$       call generate_random_blocks(0.0_wp, neffpara, nblocks, ind, break_points)
!!$
!!$       lik = likp
!!$       acptsim = 0
!!$
!!$       !$OMP PARALLEL DO SHARED(npart,ind,u,eps,acptsim,parasim,lik,prio,nblocks,i,var,T) PRIVATE(j,bvar,bj,bsize,bi,p1,p0,lik0,pr0,pr1,lik1,alp,k)
!!$       do j = 1, npart
!!$
!!$
!!$          p0 = parasim(:,j)
!!$          lik0 = lik(j)
!!$          pr0 = prio(j)
!!$
!!$
!!$          do bj = 1, nblocks
!!$             bsize = break_points(bj+1) - break_points(bj) 
!!$             allocate(bi(bsize), bvar(bsize, bsize))
!!$             bi = ind(break_points(bj)+1:break_points(bj+1))
!!$             bvar = var(bi, bi)
!!$             call dpotrf('l', bsize, bvar, bsize, info)
!!$
!!$             if (info .ne. 0) then
!!$                print*,'Cholesky failed.', info
!!$                bvar = 0.0_wp
!!$                do k = 1, bsize
!!$                   bvar(k,k) = 0.01_wp
!!$                end do
!!$
!!$             end if
!!$
!!$
!!$             do k = 1, nintmh
!!$
!!$                p1 = p0
!!$
!!$                p1(bi) = p1(bi) + matmul(bvar, eps((j-1)*nintmh + k, bi))
!!$                
!!$                pr1 = g(p1)
!!$                lik1 = f(p1,T(i))
!!$
!!$                alp = exp(lik1 - lik0 + pr1 -  pr0)
!!$
!!$                if (u((i-2)*npart*nintmh*nblocks + (j-1)*nintmh*nblocks + (bj-1)*nintmh + k) < alp) then
!!$                   p0 = p1
!!$                   lik0 = lik1
!!$                   pr0 = pr1
!!$                   acptsim(j,bi) = acptsim(j,bi) + 1
!!$                end if
!!$
!!$             end do
!!$             parasim(:,j) = p0
!!$             lik(j)  = lik0
!!$             prio(j) = pr0
!!$
!!$             deallocate(bvar, bi)
!!$          end do
!!$
!!$       end do
!!$       !$OMP END PARALLEL DO
!!$
!!$       arate(i,:) = sum(acptsim, dim=1)/(1.0_wp*npart*nintmh)
!!$
!!$       if (minval(arate(i,pmaskinv)) < 0.25) then
!!$
!!$          scale = maxval((/ scale - 0.01_wp, 0.10_wp/))
!!$       else
!!$          scale = minval((/ scale + 0.01_wp, 1.00_wp/))
!!$       end if
!!$
!!$       !------------------------------------------------------------
!!$       ! compute predictive likelihoods
!!$       !------------------------------------------------------------
!!$       if (i < nT) then
!!$          !$OMP PARALLEL DO SHARED(likp, parasim, T, i, npart) PRIVATE(j)
!!$          do j = 1, npart
!!$             likp(j) = f(parasim(:,j), T(i+1))
!!$          end do
!!$          !$OMP END PARALLEL DO
!!$       end if
!!$
!!$       !-----------------------------------------------------------
!!$       print*,''
!!$       write(*,'(a,f5.2,f5.2,f5.2)') ' acceptance rate [min,mean,max]: ', & 
!!$            minval(arate(i,pmaskinv)), sum(arate(i,pmaskinv))/neffpara, maxval(arate(i,pmaskinv))
!!$       print*, '(unweighted) average likelihood: ', sum(lik)/(npart*1.0_wp)
!!$       !------------------------------------------------------------
!!$       write(charnpara,'(i2)') npara
!!$       write(chari,'(i3.3)') T(i)
!!$
!!$       datafile = trim(basedir)//'/'//trim(chari)//'parasim.txt'
!!$       wghtfile = trim(basedir)//'/'//trim(chari)//'wtsim.txt'
!!$
!!$       open(1, file=datafile, action='write')
!!$       open(2, file=wghtfile, action='write')
!!$
!!$       do j = 1, npart
!!$          
!!$          write(1,'(41f)') parasim(:,j)
!!$          write(2,'(1f)') wtsim(j)
!!$
!!$       end do
!!$
!!$       close(1)
!!$       close(2)
!!$       write(*,*) 'Wrote files '//trim(chari)//'{parasim,wtsim,ess}.txt'
!!$       print*,''
!!$       !----------------------------------------------------------
!!$
!!$    end do
!!$    
!!$    deallocate(pmaskinv, break_points, ind)
!!$  end subroutine gewekes_smc

  subroutine systematic_resampling(npart, wtsim, randu, paraind)
    
    integer, intent(in) :: npart
    
    real(wp), intent(in) :: wtsim(npart), randu
    integer, intent(out) :: paraind(npart)

    integer :: cweights(npart), m(npart), k, cs

    cweights(1) = floor(npart*wtsim(1) + randu)
    m(1) = cweights(1)
    do k = 2, npart
       cweights(k) = floor(sum(npart*wtsim(1:k))  + randu)
    end do

    cweights(npart) = npart
    do k = 2, npart
       m(k) = cweights(k) - cweights(k-1)
    end do

    paraind = 0
    paraind(1:m(1)) = 1

    cs = m(1)

    do k = 2, npart
       paraind(cs+1:cs+m(k)) = k

       cs = cs + m(k)

    enddo

  end subroutine systematic_resampling


  subroutine multinomial_resampling(npart, wtsim, randu, paraind)

    integer, intent(in) :: npart

    real(wp), intent(in) :: wtsim(npart), randu(npart)
    integer, intent(out) :: paraind(npart)

    real(wp) :: cumsum(npart), u
    integer :: i, j


    do i = 1, npart
       cumsum(i) = sum(wtsim(1:i))
    end do

    do i = 1, npart

       u = randu(i)

       j = 1
       do 
          if (u < cumsum(j)) exit
          
          j = j + 1
       end do

       paraind(i) = j

    end do

  end subroutine multinomial_resampling



  subroutine compute_mean_and_variance(mu, var, parasim, wtsim, npara, nsim)

    integer, intent(in) :: npara, nsim
    real(wp), intent(in) :: parasim(npara, nsim), wtsim(nsim)

    real(wp), intent(out) :: mu(npara), var(npara, npara)

    real(wp) :: R(npara, npara)
    
    integer :: j

    mu = 0.0_wp
    var = 0.0_wp
    do j = 1, nsim
       mu = mu + parasim(:,j)*wtsim(j)
    end do


    do j = 1, nsim
       R = 0.0_wp
       call dger(npara, npara, 1.0_wp, parasim(:,j), 1, parasim(:, j), 1, R, npara)
       var = var + R*wtsim(j)

    end do

    call dger(npara, npara, -1.0_wp, mu, 1, mu, 1, var, npara)

    !print*,mu
  end subroutine compute_mean_and_variance



end module mcmc

