!!! prior.f90 --- 
!! 
!! Description: This module implements the prior density evaluation and
!!   generation.  Based on code by Frank Schorfheide.
!! 
!! Author: Ed Herbst [edward.p.herbst@frb.gov]
!! Last-Updated: 07/31/13
!! 

include 'mkl_vsl.fi'

module prior

  use mkl95_precision, only: wp => dp
  use mkl_vsl_type
  use mkl_vsl
  use logbeta, only : betaln, gamln


  implicit none 

  real(wp), parameter :: M_PI = 3.14159265358979_wp

contains 

  function priorrand(nsim, pshape, pmean, pstdd, pmask, pfix) result(parasim)

    integer, intent(in) :: nsim
    integer, intent(in) :: pshape(:), pmask(:)
    real(wp), intent(in) :: pmean(:), pstdd(:), pfix(:)

    real(wp), dimension(nsim, size(pshape)), target :: parasim(nsim, size(pshape))
    real(wp), dimension(:), pointer :: paraptr  

    integer i, npara, myint
    real(wp) :: a, b
    real(wp), dimension(:), allocatable :: rand_temp(:,:)

    type (VSL_STREAM_STATE) :: stream
    integer(kind=4) errcode
    integer brng, method, methodu, methodg, methodb, seed, mstorage, time_array(8), status

    integer :: i0, i1, pind, nelem, matsize, info, j, nu
    real(wp), allocatable :: psi(:,:), Z(:,:), W(:,:), work(:), rand_eps(:,:,:)
    integer, allocatable :: ipiv(:)
    
    call date_and_time(values=time_array)

    brng=VSL_BRNG_MT19937
    method=VSL_METHOD_DGAUSSIAN_ICDF
    methodu=VSL_METHOD_DUNIFORM_STD
    methodg=VSL_METHOD_DGAMMA_GNORM
    methodb=VSL_METHOD_DBETA_CJA
    mstorage=VSL_MATRIX_STORAGE_FULL

    !seed = 4211421
    seed=mod(sum(time_array),1000)
    errcode = vslnewstream(stream, brng, seed)

    npara = size(pshape)
    i = 1
    do 
       if (i > npara) exit

       if (pmask(i) == 1) then
          parasim(:, i) = pfix(i)
       else
          paraptr => parasim(:, i)
          if (pshape(i) == 1) then ! beta(a, b)
             a = (1-pmean(i))*pmean(i)**2/pstdd(i)**2 - pmean(i)
             b = a*(1/pmean(i) - 1)
             status = vdrngbeta(methodb, stream, nsim, parasim(:, i), a, b, 0.0_wp, 1.0_wp)
          elseif (pshape(i) == 2) then ! gamma(a, b)
             b = pstdd(i)**2/pmean(i); !theta
             a = pmean(i)/b;    
             status = vdrnggamma(methodg, stream, nsim, parasim(:, i), a, 0.0_wp, b)
          elseif (pshape(i) == 3) then  !normal(a, b)
             a = pmean(i)
             b = pstdd(i)
             status = vdrnggaussian(method, stream, nsim, parasim(:, i), a, b)
          elseif (pshape(i) == 4) then ! inv gamma
             a = pmean(i)
             b = pstdd(i)
             allocate(rand_temp(int(b), nsim)) 
             status = vdrnggaussian(method, stream, nsim*int(b), rand_temp, 0.0_wp, 1.0_wp)
             parasim(:, i) = sqrt(b*a**2 / sum(rand_temp**2, 1))
             deallocate(rand_temp)
          elseif (pshape(i) == 5) then ! uniform
             a = pmean(i)
             b = pstdd(i)
             status = vdrnguniform(methodu, stream, nsim, parasim(:, i), a, b)
          elseif (pshape(i) == 0) then
             a = pfix(i)
             paraptr = a
          elseif (pshape(i) > 90) then

             pind = i 
             i0 = i 
             i = i + 1
             do 
                if (pshape(i) .ne. pshape(pind)) exit
                i = i + 1
             end do
             i = i - 1
             i1 = i

             nelem = i1 - i0 + 1
             matsize = int((sqrt(1.0_wp+8.0_wp*nelem)-1.0_wp)/2.0_wp)

             nu = int(pstdd(pind))
             allocate(psi(matsize,matsize),Z(matsize,nu),W(matsize,matsize), &
                  ipiv(matsize), work(matsize))

             psi = diag_first_ivec(pmean(i0:i1),matsize,nelem)
             allocate(rand_eps(matsize,nu,nsim))
             status = vdrnggaussian(method, stream, nsim*nu*matsize, rand_eps, 0.0_wp, 1.0_wp)

             call dgetrf(matsize, matsize, psi, matsize, ipiv, info)
             call dgetri(matsize, psi, matsize, ipiv, work, matsize, info)


             call dpotrf('l', matsize, psi, matsize, info)


             do j = 1,nsim
                call dsymm('l','l',matsize,nu,1.0_wp,psi,matsize,rand_eps(:,:,j),matsize,0.0_wp, Z, matsize)
                call dgemm('n','t',matsize,matsize,nu,1.0_wp,Z,matsize,Z,matsize,0.0_wp,W,matsize)
                call dgetrf(matsize, matsize, W, matsize, ipiv, info)
                call dgetri(matsize, W, matsize, ipiv, work, matsize, info)
                parasim(j,i0:i1) = diag_first_vec(W,matsize,nelem)

             end do
             deallocate(psi,Z,W,rand_eps,ipiv,work)
          end if
       end if

       i = i + 1
    end do

  end function priorrand

  function diag_first_vec(x,n,vn) result(vecx)

    integer, intent(in) :: n, vn
    real(wp), intent(in) :: x(n,n)

    real(wp) :: vecx(vn)

    integer :: i,j,vi

    vi = 1
    do i = 1,n
       vecx(vi) = x(i,i)
       vi = vi+1
    end do

    do i = 1,n-1
       do j = i+1,n
          vecx(vi) = x(i,j)
          vi = vi+1
       end do
    end do

  end function diag_first_vec

  function diag_first_ivec(vecx,n,vn) result(x)

    integer, intent(in) :: n, vn
    real(wp), intent(in) :: vecx(vn)
    real(wp) :: x(n,n)

    integer :: i, j, vi

    vi = 1
    do i = 1,n
       x(i,i) = vecx(vi)
       vi = vi+1
    end do

    do i = 1,n-1
       do j=i+1,n
          x(i,j) = vecx(vi)
          x(j,i) = x(i,j)
          vi = vi+1
       end do
    end do

  end function diag_first_ivec

  function priordens(para, pmean, pstdd, pshape, pmask, pfix) result(logprior)
    ! returns the log prior density of a vector of parameters
    integer, intent(in) :: pshape(:), pmask(:)
    real(wp), intent(in) :: pmean(:), pstdd(:), pfix(:), para(:)

    real(wp), allocatable :: psi, randiw

    real(wp) :: logprior, a, b, pr0
    integer :: npara, i, i0, i1, pind, nelem, matsize, info, nu 

    real(wp), allocatable :: S(:,:), X(:,:)

    logprior = 0.0_wp
    npara = size(para, 1)

    i = 1
    ! do error checking...
    do 
       if (i > npara) exit
       
       if (pmask(i) == 1) then

       else
          pr0 = logprior

          if (pshape(i) == 1) then ! beta(a, b)
             a = (1-pmean(i))*pmean(i)**2/pstdd(i)**2 - pmean(i)
             b = a*(1/pmean(i) - 1)
             logprior = logprior + logbetapdf(para(i),a,b)
             !             print*,i,logbetapdf(para(i),a,b)
          elseif (pshape(i) == 2) then ! gamma(a, b)
             b = pstdd(i)**2/pmean(i); !theta
             a = pmean(i)/b;           !k
             logprior = logprior + loggampdf(para(i),a,b)
             !            print*,i,loggampdf(para(i),a,b)
          elseif (pshape(i) == 3) then  !normal(a, b)
             a = pmean(i)
             b = pstdd(i)
             logprior = logprior + lognorpdf(para(i),a,b)
             !             print*,i,lognorpdf(para(i),a,b)
          elseif (pshape(i) == 4) then ! inv gamma
             a = pmean(i)
             b = pstdd(i)
             logprior = logprior + logigpdf(para(i),a,b)
             !             print*,i,logigpdf(para(i),a,b)
          elseif (pshape(i) == 5) then ! uniform
             a = pmean(i)
             b = pstdd(i)
             logprior = logprior + log(1.0_wp/(b - a))
             !             print*,i,log(1.0_wp/(b-a))
          elseif (pshape(i) == 0) then
             a = pfix(i)
          elseif (pshape(i) > 90) then ! inverse wishart...
             pind = i 
             i0 = i 
             i = i + 1
             do 
                if (pshape(i) .ne. pshape(pind)) exit
                i = i + 1
             end do
             i = i - 1

             i1 = i

             nelem = i1 - i0 + 1
             matsize = int((sqrt(1.0_wp+8.0_wp*nelem)-1.0_wp)/2.0_wp)
             nu = int(pstdd(pind))
             allocate(S(matsize,matsize),X(matsize,matsize))
             S = diag_first_ivec(pmean(i0:i1),matsize,nelem)
             X = diag_first_ivec(para(i0:i1),matsize,nelem)
             logprior = logprior + logiwishpdf(nu,S,X,matsize)            
             deallocate(S,X)

          end if

       end if


       i = i + 1
    end do
 




  end function priordens

  function logbetapdf(x, a, b)

    real(wp), intent(in) :: x, a, b
    real(wp) :: logbetapdf

    logbetapdf = -betaln(a,b) + (a-1.0_wp)*log(x) + (b-1.0_wp)*log(1.0-x)
    
  end function logbetapdf

  function loggampdf(x, a, b)

    real(wp), intent(in) :: x, a, b
    real(wp) :: loggampdf

    loggampdf = -gamln(a) -a*log(b) + (a-1.0_wp)*log(x) - x/b
    
  end function loggampdf

  function logiwishpdf(nu,S,X,n)

    integer, intent(in) :: nu, n
    real(wp), intent(in) :: S(n,n), X(n,n)
    real(wp) :: logiwishpdf
    integer :: i, ipiv(n), info
    real(wp) :: kap, iX(n,n), tr, work(n), ret(n,n)

    iX = X
    call dgetrf(n, n, iX, n, ipiv, info)
    call dgetri(n, iX, n, ipiv, work, n, info)
    call dgemm('n','n',n,n,n,1.0_wp,iX,n,S,n,0.0_wp,ret,n)

    tr = 0.0_wp
    kap = 0.0_wp
    do i = 1,n
       tr = tr + ret(i,i)
       kap = kap + gamln(0.5*(nu*1.0_wp-i*1.0_wp-1.0_wp))
    end do

    kap = -kap - 0.5_wp*nu*n*log(2.0_wp) - 0.25_wp*n*(n-1)*log(M_PI)
    logiwishpdf = kap + 0.5_wp*nu*log(determinant(S,n)) & 
         - 0.5_wp*(nu+n+1)*log(determinant(X,n)) -0.5_wp*tr

  end function logiwishpdf
  function lognorpdf(x, a, b)

    real(wp), intent(in) :: x, a, b
    real(wp) :: lognorpdf

    lognorpdf = -0.5_wp*log(2.0_wp*M_PI) - log(b) - 0.5_wp*(x-a)**2/b**2

  end function lognorpdf

  function logigpdf(x,a,b)

    real(wp), intent(in) :: x, a, b
    real(wp) :: logigpdf

    logigpdf = log(2.0_wp) - gamln(b/2.0_wp) + b/2.0_wp*log(b*a**2/2.0_wp) &
         -(b+1.0_wp)/2.0_wp*log(x**2) - b*a**2.0/(2.0_wp*x**2)

  end function logigpdf



  function dlpdfbeta(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: dlpdfbeta
    dlpdfbeta = (a-1.0_wp)/x - (b-1.0_wp)/(1.0_wp-x);
  endfunction dlpdfbeta

  function dlpdfgam(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: dlpdfgam
    dlpdfgam = (a-1.0_wp)/x - 1.0_wp/b;
  endfunction dlpdfgam

  function dlpdfnor(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: dlpdfnor
    dlpdfnor =  -(x-a)/b**2;
  endfunction dlpdfnor

  function dlpdfig(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: dlpdfig
    dlpdfig = -(b+1.0_wp)/x + b*a**2/x**3;
  endfunction dlpdfig


  function dlpriordens(para, pmean, pstdd, pshape, pmask, pfix) result(dlnprior)
    ! returns the log prior density of a vector of parameters

    integer, intent(in) :: pshape(:), pmask(:)
    real(wp), intent(in) :: pmean(:), pstdd(:), pfix(:), para(:)

    real(wp) :: a, b
    real(wp) :: dlnprior(size(pshape,1))

    integer :: npara, i


    npara = size(para, 1)
    dlnprior = 0.0_wp
    ! do error checking...
    do i = 1, npara

       if (pmask(i) == 1) then

       else

          if (pshape(i) == 1) then ! beta(a, b)
             a = (1-pmean(i))*pmean(i)**2/pstdd(i)**2 - pmean(i)
             b = a*(1/pmean(i) - 1)
             dlnprior(i) = dlpdfbeta(para(i),a,b)
          elseif (pshape(i) == 2) then ! gamma(a, b)
             b = pstdd(i)**2/pmean(i); !theta
             a = pmean(i)/b;           !k
             dlnprior(i) = dlpdfgam(para(i),a,b)
          elseif (pshape(i) == 3) then  !normal(a, b)
             a = pmean(i)
             b = pstdd(i)
             dlnprior(i) = dlpdfnor(para(i),a,b)
          elseif (pshape(i) == 4) then ! inv gamma
             a = pmean(i)
             b = pstdd(i)
             dlnprior(i) = dlpdfig(para(i),a,b)
          elseif (pshape(i) == 5) then ! uniform
             a = pmean(i)
             b = pstdd(i)
             dlnprior(i) = 0.0_wp
          elseif (pshape(i) == 0) then
             a = pfix(i)
          end if
       end if

    end do


  end function dlpriordens


  function hlpdfbeta(x,a,b)
    real(wp), intent(in) :: x, a, b
    real(wp) :: hlpdfbeta

    hlpdfbeta  = -(a-1.0_wp)/(x**2) - (b-1.0_wp)/( (1.0_wp-x)**2);
  endfunction hlpdfbeta

  function hlpdfgam(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: hlpdfgam

    hlpdfgam = -(a-1.0_wp)/(x**2);
  endfunction hlpdfgam

  function hlpdfnor(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: hlpdfnor

    hlpdfnor = -1.0_wp/b**2;
  endfunction hlpdfnor

  function hlpdfig(x,a,b);
    real(wp), intent(in) :: x, a, b
    real(wp) :: hlpdfig

    hlpdfig = (b+1.0_wp)/x**2 -3.0_wp*b*a**2/x**4;
  endfunction hlpdfig

  function hlpriordens(para, pmean, pstdd, pshape, pmask, pfix) result(hlnprior)
    ! returns the log prior density of a vector of parameters

    integer, intent(in) :: pshape(:), pmask(:)
    real(wp), intent(in) :: pmean(:), pstdd(:), pfix(:), para(:)

    real(wp) :: a, b
    real(wp) :: hlnprior(size(pshape,1),size(pshape,1))

    integer :: npara, i


    npara = size(para, 1)
    hlnprior = 0.0_wp
    ! do error checking...
    do i = 1, npara

       if (pmask(i) == 1) then

       else

          if (pshape(i) == 1) then ! beta(a, b)
             a = (1-pmean(i))*pmean(i)**2/pstdd(i)**2 - pmean(i)
             b = a*(1/pmean(i) - 1)
             hlnprior(i, i) = hlpdfbeta(para(i),a,b)
          elseif (pshape(i) == 2) then ! gamma(a, b)
             b = pstdd(i)**2/pmean(i); !theta
             a = pmean(i)/b;           !k
             hlnprior(i, i) = hlpdfgam(para(i),a,b)
          elseif (pshape(i) == 3) then  !normal(a, b)
             a = pmean(i)
             b = pstdd(i)
             hlnprior(i, i) = hlpdfnor(para(i),a,b)
          elseif (pshape(i) == 4) then ! inv gamma
             a = pmean(i)
             b = pstdd(i)
             hlnprior(i, i) = hlpdfig(para(i),a,b)
          elseif (pshape(i) == 5) then ! uniform
             a = pmean(i)
             b = pstdd(i)
             hlnprior(i, i) = 0.0_wp
          elseif (pshape(i) == 0) then
             a = pfix(i)
          end if
       end if

    end do


  end function hlpriordens

  function determinant(matrix, r) result(det)
    ! Computes the determinant of symmetric square matrix, matrix (rank r).
    integer, intent(in) :: r
    real(wp), intent(in) :: matrix(r,r)

    real(wp) :: det

    integer :: info, i, piv(r)
    real(wp) :: matrix_copy(r, r)

    call dcopy(r*r, matrix, 1, matrix_copy, 1)
    det = 0.0_wp

    call dpotrf('u', r, matrix_copy, r, info)

    if (info .ne. 0) then
       !write(*,'(a,i4)') 'In determinant(), dgetrf returned error code ', info
       det = -10000.0_wp
       return
    end if

    det = 1.0_wp

    do i = 1, r

       if (.true.) then !(piv(i) .ne. i) then
          det = det * matrix_copy(i, i) * matrix_copy(i,i)
       else
          det = det * matrix_copy(i, i)
       end if

    end do
  end function determinant


end module prior
