!!! rand.f90 --- 
!! 
!! Description: A collection of routines for random variables.
!! 
!!  Contains:
!!
!!    function mvt_pdf
!!    subroutn vdrngmvt
!!    function mvnormal_pdf
!! 
!! Author: Ed Herbst [edward.p.herbst@frb.gov]
!! Last-Updated: 11/16/12
!! 
module rand

  use mkl95_precision, only: wp => dp
  use mkl_vsl_type
  use mkl_vsl


  use logbeta, only : betaln, gamln

contains


  function mvt_pdf(x, mu, chol_sigma, df) result(logq)
    !! Computes the log of the n-dimensional multivariate t pdf at x
    !! with mean mu and variance = chol_sigma*chol_sigma' and dof = df
    !!

    integer, intent(in) :: df
    real(wp), intent(in) :: x(:), mu(:), chol_sigma(:,:)
    real(wp), external :: ddot

    real(wp) :: logq
    real(wp) :: a(size(x, 1)), det_sigma

    integer :: n, i
    n = size(x, 1)

    if (n /= size(mu, 1)) then
       print*, 'mvnormal pdf, size error'
       stop
    endif

    det_sigma = 1.0_wp

    do i = 1, n

       det_sigma = det_sigma*chol_sigma(i,i)

    end do
    det_sigma = det_sigma**2
    a = x - mu
    call dtrsv('l','n', 'n', n, chol_sigma, n, a, 1)

    logq = gamln(0.5_wp*(n + df)) - 0.5_wp*log(det_sigma) - 0.5_wp*n*log(df*3.14159265358979_wp) &
         - gamln(0.5_wp*df) - 0.5_wp*(df+n)*log(1.0_wp + ddot(n, a, 1, a, 1)/(df*1.0_wp))

  end function mvt_pdf

  subroutine vdrngmvt(stream, n, deviates, mean, cholvar, df)
    !! generates multivariate t random 
    !!
    !!
    integer, intent(in) :: n, df
    type (VSL_STREAM_STATE), intent(inout) :: stream

    real(wp), intent(in) :: mean(n), cholvar(n,n)
    real(wp), intent(out) :: deviates(n)

    real(wp), external :: ddot
    real(wp) :: normdraw(n), nu(1)

    integer :: status, method, i

    method = VSL_METHOD_DGAUSSIAN_ICDF
    status = vdrnggaussian(method, stream, n, normdraw, 0.0_wp, 1.0_wp)

!    call dgemv('n', n, n, 1.0_wp, cholvar, n, normdraw, 1, 0.0_wp, deviates, 1)
    !call dsymv('l', n, 1.0_wp, transpose(cholvar), n, normdraw, 1, 0.0_wp, deviates, 1)

    do i = 1,n
       deviates(i) = ddot(i,cholvar(i,1:i), 1,normdraw(1:i),1)
    end do
    !deviates = mean + matmul(cholvar, normdraw)

    method = VSL_METHOD_DGAMMA_GNORM
    status = vdrnggamma(method, stream, 1, nu, (1.0_wp*df) / 2.0_wp, 0.0_wp, 1.0_wp)

    nu = 2.0_wp*nu / (df*1.0_wp)

    normdraw = mean

    call daxpy(n, 1.0_wp/sqrt(nu), deviates, 1, normdraw, 1)

    deviates = normdraw

  end subroutine vdrngmvt


  function mvnormal_pdf(x, mu, chol_sigma) result(logq)
    !! Computes the log of the n-dimensional multivariate normal pdf at x
    !! with mean mu and variance = chol_sigma*chol_sigma'.
    !!
    real(wp), intent(in) :: x(:), mu(:), chol_sigma(:,:)
    real(wp), external :: ddot

    real(wp) :: logq
    real(wp) :: a(size(x, 1)), det_sigma

    integer :: n, i
    n = size(x, 1)

    if (n /= size(mu, 1)) then
       print*, 'mvnormal pdf, size error'
       stop
    endif

    det_sigma = 1.0_wp

    do i = 1, n

       det_sigma = det_sigma*chol_sigma(i,i)

    end do
    det_sigma = det_sigma**2
    a = x - mu
    call dtrsv('l','n', 'n', n, chol_sigma, n, a, 1)

    logq = -n*0.5_wp*log(2.0_wp*3.14159_wp) - 0.5_wp*log(det_sigma) - 0.5*ddot(n, a, 1, a, 1)

  end function mvnormal_pdf
  

end module rand
